# -*- coding: utf-8 -*-
"""
Datatypes and parsers for the various structures, specific to Redis Gears.

These datatypes are returend from various redgrease functions, merely for the purpose
of providig more convenient structure, typing and documentation compared to the native
'list-based' structures.

They are generally not intended to be instantiated by end-users.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import ast
import sys
from typing import Any, Dict, Generic, Iterable, List, Optional, TypeVar, Union

import attr
import cloudpickle
import wrapt

import redgrease.gears
from redgrease.utils import (
    REnum,
    bool_ok,
    list_parser,
    safe_bool,
    safe_str,
    str_if_bytes,
    to_bytes,
    to_dict,
    to_kwargs,
)

T = TypeVar("T")


@attr.s(frozen=True, auto_attribs=True)
class ExecID:
    """Execution ID

    Attributes:
        shard_id (str):
            Shard Identifier

        sequence (in):
            Sequence number
    """

    shard_id: str = "0000000000000000000000000000000000000000"
    sequence: int = 0

    def __str__(self):
        return f"{self.shard_id}-{self.sequence}"

    def __bytes__(self):
        return str(self).encode()

    def __repr__(self):
        class_name = self.__class__.__name__
        return (
            f"{class_name}(" f"shard_id={self.shard_id}," f"sequence={self.sequence})"
        )

    @staticmethod
    def parse(value: Union[str, bytes]) -> "ExecID":
        """Parses a string or bytes representation into a `redgrease.data.ExecID`

        Returns:
            redgrease.data.ExecID:
                Theo parsed ExecID

        Raises:
            ValueError:
                If the the value cannot be parsed.
        """
        if isinstance(value, bytes):
            value = value.decode()

        try:
            values = value.split("-")

            shard_id = values[0]
            sequence = int(values[1])
        except (AttributeError, IndexError) as e:
            raise ValueError(
                f"Unable to parse ExecID. Invalid serialization: '{value}'"
            ) from e

        return ExecID(shard_id=shard_id, sequence=sequence)


class ExecutionResult(wrapt.ObjectProxy, Generic[T]):
    """Common class for all types of execution results.
    Generic / Polymorphic on the result type (T) of the Gears function.

    Redis Gears specifies a few different commands for getting the results of a
    function execution (`pyexecute`, `getresults`, `getresultsblocking` and `trigger`),
    each potentially having more than one different possible value type, depending on
    context.

    In addition, while most gears functions result in collection of values, some
    functions (notably those ending with a `count` or `avg` operation) semantically
    always have scalar results, but are still natively returned as a list.

    redgrease.data.ExecutionResult is a unified result type for all scenarios,
    aiming at providing as intuitive API experience as possible.

    It is generic and walks and quacks just like the main result type (T) it wraps.
    With some notable exceptions:
        - It has an additional property `errors` containing any acumulated errors
        - Scalar results, behaves like scalars, but **also** like a single element list.

    This behavior isn't always perfect but gives for the most part an intuitive api
    experience.

    If the behaviour in some situations are confusing, the raw wrapped value can be
    accessed through the `value` property.
    """

    def __init__(self, value: T, errors: Optional[List] = None):
        wrapt.ObjectProxy.__init__(self, value)
        self._self_errors: Optional[List] = errors

    @property
    def value(self):
        """Gets the raw result value of the Gears function execution.

        Returns:
            T:
                The result value / values
        """
        return self.__wrapped__

    @property
    def errors(self) -> Optional[List]:
        return self._self_errors

    def __repr__(self) -> str:
        cls_nm = "ExecutionResult"
        val_typ_nm = self.__class__.__name__
        if self._self_errors:
            err = f", errors={repr(self._self_errors)}"
        else:
            err = ""

        return f"{cls_nm}[{val_typ_nm}]({self.__wrapped__.__repr__()}{err})"

    def __iter__(self):
        if hasattr(self.value, "__iter__"):
            return iter(self.value)
        else:
            return iter([] if self.value is None else [self.value])

    def __len__(self) -> int:
        if hasattr(self.value, "__len__") and not isinstance(self.value, (str, bytes)):
            return len(self.value)
        else:
            return 0 if self.value is None else 1

    def __eq__(self, other):
        if not isinstance(self.value, list) and isinstance(other, list):
            return [self.value] == other
        else:
            return self.value == other

    def __getitem__(self, *args, **kwargs):
        if hasattr(self.value, "__getitem__"):
            return self.value.__getitem__(*args, **kwargs)
        else:
            if args and args[0] == 0:
                return self.value
            else:
                raise TypeError(f"{self} is not subscriptable")

    def __contains__(self, val) -> bool:
        if hasattr(self.value, "__contains__"):
            return val in self.value
        else:
            return self.value == val

    def __bytes__(self) -> bytes:
        return to_bytes(self.value)


def parse_execute_response(response, pickled=False) -> ExecutionResult:
    """Parses raw responses from `pyexecute`, `getresults` and `getresultsblocking`
    into a `redgrease.data.ExecuteResponse` object.

    Args:
        response (Any):
            The raw gears function response.
            This is most commonly a tuple of a list with the actual results and a list
            of errors  List[List[Union[T, Any]]].

            For some scenarios the response may take other forms, like a simple `Ok`
            (e.g. in the absence of a closing `run()` operation) or an excecution ID
            (e.g. for non-blocking executions).

        pickled (bool, optional):
            Indicates if the response is pickled and need to be unpickled.
            Defaults to False.

    Returns:
        ExecutionResult[T]:
            A parsed execution response
    """
    if bool_ok(response):
        # Any "Ok" response is just True
        return ExecutionResult(True)
    elif isinstance(response, list) and len(response) == 2:
        # The 'normal' list-of-lists response
        # Results are unpickled if they are pickled
        # Note that the special case when the resiult only has one element,
        #   then the single value is used instread
        #   as the ExecutionResults would pretend it is a list anyway, if needed
        #   This way scalar results from for example `count` and `avg` will behave
        #   like scalars which is what is typically wanted

        result, errors = response
        if isinstance(result, list):
            if pickled:
                result = [cloudpickle.loads(value) for value in result]
            if len(result) == 1:
                result = result[0]
        return ExecutionResult(result, errors=errors)
    elif isinstance(response, bytes):
        # Bytes response means ExecID
        return ExecutionResult(ExecID.parse(response))
    else:
        # If the response doesnt fit any known pattern, its returned as-is
        return ExecutionResult(response)


def parse_trigger_response(response, pickled=False) -> ExecutionResult:
    """Parses raw responses from `trigger` into a `redgrease.data.ExecuteResponse`
    object.

    Args:
        response (Any):
            The gears function response.
            This is a tuple of a list with the actual results and a list of errors
            List[List[Union[T, Any]]].

        pickled (bool, optional):
            Indicates if the response is pickled and need to be unpickled.
            Defaults to False.

    Returns:
        ExecutionResult[T]:
            A parsed execution response
    """
    if pickled:
        response = [cloudpickle.loads(value) for value in response]

    if len(response) == 1:
        response = response[0]
    return ExecutionResult(response)


class ExecutionStatus(REnum):
    """Representation of the various states an execution could be in."""

    created = b"created"
    """Created - The execution has been created."""

    running = b"running"
    """Running - The execution is running."""

    done = b"done"
    """Done - Thee execution is done."""

    aborted = b"aborted"
    """Aborted - The execution has been aborted."""

    pending_cluster = b"pending_cluster"
    """Pending Cluster - Initiator is waiting for all workers to finish."""

    pending_run = b"pending_run"
    """Pending Run - Worker is pending ok from initiator to execute."""

    pending_receive = b"pending_receive"
    """Pending Receive - Initiator is pending acknowledgement from workers on receiving
    execution.
    """

    pending_termination = b"pending_termination"
    """Pending Termination - Worker is pending termination messaging from initiator"""


# TODO: Isn't this sugar rather than data?
class ExecLocality(REnum):
    """Locality of exetution: Shard or Cluster"""

    Shard = "Shard"
    Cluster = "Cluster"


class RedisObject:
    """Base class for many of the more complex Redis Gears configuration values"""

    @classmethod
    def from_redis(cls, params) -> "RedisObject":
        """Default parser

        Assumes the object is seralized as a list of alternating attribute names and
        values.

        Note: This method should not be invoked directly on the 'RedisObject'
        base class.
        It should be only be invoked on subclasses of RedisObjects.

        Returns:
            RedisObect:
                Returns the RedisObject subclass if, and only if, its constructor
                argument names and value types exactly match the names and values
                in the input list.

        Raises:
            TypeError:
                If either the input list contains attributes not defined in the
                subclass consructor, or if the subclass defines mandatory constructor
                arguments that are not present in the input list.
        """
        return cls(**to_kwargs(params))  # type: ignore


def parse_PD(value: Union[str, bytes]) -> Dict:
    """Parses str or bytes to a dict.

    Used for the 'PD' field in the 'Registration' type, returned by 'getregistrations'.

    Args:
        value (Union[str,bytes]):
            Serialized version of a dict.

    Returns:
        Dict:
            A dictionary
    """
    str_val = safe_str(value)
    return dict(ast.literal_eval(str_val))


@attr.s(auto_attribs=True, frozen=True)
class ExecutionInfo(RedisObject):
    """Return object for `redgrease.client.Redis.dumpexecutions` command."""

    executionId: ExecID = attr.ib(converter=ExecID.parse)  # type: ignore #7912
    """The execution Id"""

    status: ExecutionStatus = attr.ib(converter=ExecutionStatus)
    """The status"""

    registered: bool
    """Indicates whether this is a registered execution"""


@attr.s(auto_attribs=True, frozen=True)
class RegData(RedisObject):
    """Object reprenting the values for the `Registration.RegistrationData`, part of
    the return value of `redgrease.client.dupregistrations` command.
    """

    mode: str = attr.ib(converter=safe_str)
    """Registration mode."""

    numTriggered: int
    """A counter of triggered executions."""

    numSuccess: int
    """A counter of successful executions."""

    numFailures: int
    """A counter of failed executions."""

    numAborted: int
    """A conter of aborted executions."""

    lastError: str
    """The last error returned."""

    args: Dict[str, Any] = attr.ib(converter=to_kwargs)
    """Reader-specific arguments"""


# @dataclass
@attr.s(auto_attribs=True, frozen=True)
class Registration(RedisObject):
    """Return object for `redgrease.client.Redis.dumpregistrations` command.
    Contains the information about a function registration.

    """

    id: ExecID = attr.ib(converter=ExecID.parse)  # type: ignore #7912
    """The registration ID."""

    reader: str = attr.ib(converter=safe_str)
    """The Reader."""

    desc: str
    """The description."""

    RegistrationData: RegData = attr.ib(
        converter=RegData.from_redis  # type: ignore #7912
    )
    """Registration Data, see `RegData`."""

    PD: Dict[Any, Any] = attr.ib(converter=parse_PD)
    """Private data"""


@attr.s(auto_attribs=True, frozen=True)
class ExecutionStep(RedisObject):
    """Object reprenting a 'step' in the `ExecutionPlan.steps`, attribut of
    the return value of `redgrease.client.getexecution` command.
    """

    type: str = attr.ib(converter=safe_str)
    """Step type."""

    duration: int
    """The step's duration in milliseconds (0 when `ProfileExecutions` is disabled)"""

    name: str = attr.ib(converter=safe_str)
    """Step callback"""

    arg: str = attr.ib(converter=safe_str)
    """Step argument"""


@attr.s(auto_attribs=True, frozen=True)
class ExecutionPlan(RedisObject):
    """Object representing the exetution plan for a given shard in the response from the
    `redgrease.client.Redis.getexetution` command.
    """

    status: ExecutionStatus = attr.ib(converter=ExecutionStatus)
    """The current status of the execution."""

    shards_received: int
    """Number of shards that received the execution."""

    shards_completed: int
    """Number of shards that completed the execution."""

    results: int
    """Count of results returned."""

    errors: int
    """Count of the errors raised."""

    total_duration: int
    """Total execution duration in milliseconds."""

    read_duration: int
    """Reader execution duration in milliseconds."""

    steps: List[ExecutionStep] = attr.ib(
        converter=list_parser(ExecutionStep.from_redis)  # type: ignore #7912
    )
    """The steps of the execution plan."""

    @staticmethod
    def parse(res: Iterable[bytes]) -> Dict[str, "ExecutionPlan"]:
        """Parse the raw results of `redgrease.client.Redis.getexetution` into a dict
        that maps from shard identifiers to ExecutionStep objects.

        Returns:
            Dict[str, ExecutionPlan]:
                Execution plan mapping.
        """
        return {
            str_if_bytes(shard[b"shard_id"]): ExecutionPlan.from_redis(  # type: ignore
                shard[b"execution_plan"]  # type: ignore
            )
            for shard in map(to_dict, res)
        }


@attr.s(auto_attribs=True, frozen=True)
class ShardInfo(RedisObject):
    """Object representing a shard in the `CluserInfo.shards` attribute in the response
    from `redgrease.client.Redis.infocluster` command.
    """

    id: str = attr.ib(converter=safe_str)
    """The shard's identifoer int the cluster."""

    ip: str = attr.ib(converter=safe_str)
    """The shard's IP address."""

    port: int
    """The shard's port."""

    unixSocket: str = attr.ib(converter=safe_str)
    """The shards UDS."""

    runid: str = attr.ib(converter=safe_str)
    """The engines run identifier."""

    minHslot: int
    """Lowest hash slot served by the shard."""

    maxHslot: int
    """Highest hash slot served by the shard."""


@attr.s(auto_attribs=True, frozen=True)
class ClusterInfo(RedisObject):
    """Information about the Redis Gears cluster.

    Return object for `redgrease.client.Redis.infocluster` command.
    """

    my_id: str
    """The identifier of the shard the client is connected to."""

    shards: List[ShardInfo] = attr.ib(
        converter=list_parser(ShardInfo.from_redis)  # type: ignore #7912
    )
    """List of the all the shards in the cluster."""

    @staticmethod
    def parse(res: bytes) -> Optional["ClusterInfo"]:
        """Parses the response from `redgrease.client.Redis.infocluster` into a
        `ClusterInfo` object.

        If the client is not connected to a Redis Gears cluster, `None` is returned.

        Returns:
            Optional[ClusterInfo]:
                A ClusterInfo object or None (if not in cluster mode).s
        """

        if not res or res == b"no cluster mode":
            return None

        cluster_info = ClusterInfo(
            my_id=safe_str(res[1]),
            shards=res[2],
        )

        return cluster_info


@attr.s(auto_attribs=True, frozen=True)
class PyStats(RedisObject):
    """Memory usage statistics from the Python interpreter.
    As returned by `redgrease.client.Redis.pystats`
    """

    TotalAllocated: int
    """A total of all allocations over time, in bytes."""

    PeakAllocated: int
    """The peak allocations, in bytes."""

    CurrAllocated: int
    """The currently allocated memory, in bytes."""


@attr.s(auto_attribs=True, frozen=True)
class PyRequirementInfo(RedisObject):
    """Information about a Python requirement / dependency."""

    GearReqVersion: int
    """An internally-assigned version of the requirement.
    (note: this isn't the package's version)
    """

    Name: str = attr.ib(converter=safe_str)
    """The name of the requirement as it was given to the 'requirements' argument
    of the `pyexecute` command.
    """

    IsDownloaded: bool = attr.ib(converter=safe_bool)
    """`True` if the requirement wheels was successfully download, otherwise `False`.
    """

    IsInstalled: bool = attr.ib(converter=safe_bool)
    """`True` if the requirement wheels was successfully installed, otherwise `False`.
    """

    CompiledOs: str = attr.ib(converter=safe_str)
    """The underlying Operating System"""

    Wheels: List[str] = attr.ib(
        converter=lambda wheels: safe_str(wheels)  # type: ignore
        if isinstance(wheels, bytes)
        else [safe_str(wheel) for wheel in wheels]
    )
    """A List of Wheels required by the requirement."""


def deseralize_gear_function(
    serialized_gear: str, python_version: str
) -> redgrease.gears.GearFunction:
    """Safely deserializes (unpickles) a serialized GearFunction.

    This function is only executed on the Gear server.

    It handles and appropriately reports errors due to mismatch between
    client and server Python versions.

    Args:
        serialized_gear (str):
            The serialized (cloudpickled) GearFunction.

        python_version (str):
            The python version of the client.

    Raises:
        SystemError:
            If the Python versions of the client and server mismatch.

    Returns:
        redgrease.gears.GearFunction:
            The Gear function, as created on at the client.
    """
    try:
        return cloudpickle.loads(serialized_gear)
    except Exception as err:
        import sys

        def pystr(pyver):
            return "Python %s.%s" % pyver

        runtime_python_version = sys.version_info[:2]
        function_python_version = ast.literal_eval(str(python_version))[:2]
        if runtime_python_version != function_python_version:
            raise SystemError(
                "Client / server Python version mismatch."
                f"{pystr(runtime_python_version)} runtime unable to execute "
                f"Gears functions created in {function_python_version}. "
            ) from err
        raise


def seralize_gear_function(gear_function: redgrease.gears.ClosedGearFunction) -> str:
    """Serializes a GearFunction into a wrapper code-string that can be sent to the
    Gear server to execute.

    Args:
        gear_function (redgrease.gears.ClosedGearFunction):
            GearFunction to serialize.

    Returns:
        str:
            Code string that will execute the GearFunction on the server.
    """

    # The Gear function is serialized with 'cloudpickle' and embedded in a
    # code string that will de-serialize it back and then 'construct' the actual
    # Gear function and run it.

    return f"""
import redgrease.data
import redgrease.runtime
gear_function = redgrease.data.deseralize_gear_function(
    {cloudpickle.dumps(gear_function, protocol=4)},
    python_version={tuple(sys.version_info)},
)
redgrease.runtime.run(gear_function, GearsBuilder)
"""



# -*- coding: utf-8 -*-
"""
Module for handling of requirements.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import itertools
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Union

import packaging.requirements
import packaging.version

default_redgrease = packaging.requirements.Requirement("redgrease[runtime]")
"""The default redgrease runtime package. Latest from PyPi."""

# Re-exposing for convenience
Version = packaging.version.Version
"""Same as 'packaging.version.Version', for convenience."""

Requirement = packaging.requirements.Requirement
"""Same as 'packaging.requirements.Requirement', for convenience."""

InvalidRequirement = packaging.requirements.InvalidRequirement
"""Same as 'packaging.requirements.InvalidRequirement', for convenience."""

PackageOption = Optional[Union[bool, str, Version, Requirement]]
"""Type alias for valid types of RedGrease package version specification."""


def read_requirements(
    *requirements_file_paths: str,
) -> Set[Requirement]:
    """Extract list of individual requirements from a 'requirements.txt' file

    Args:
        requirements_file_paths (Iterable[str]):
            Requirements file path

    Returns:
        Set[str]:
            The individual requirements in the file.
    """
    requirements = set.union(
        *[
            set(
                map(
                    lambda req_line: Requirement(req_line.split("#")[0]),
                    filter(
                        lambda line: len(line) > 0 and not line.startswith("#"),
                        map(str.strip, open(file_path).readlines()),
                    ),
                )
            )
            for file_path in requirements_file_paths
        ]
    )

    return resolve_requirements(requirements)


def to_packages_dict(
    requirements: Iterable[Requirement],
) -> Dict[str, List[Requirement]]:
    """Transforms a collection of requirements to a dict mapping packge names to
    a list of requirements.

    I.e. grouping the different versions of packages in the requiremnts by name.

    E.g.
        ```
        reqs = map(Requirement, ["pack_x[xtra]", "pack_y>=1.0", "pack_x!=2.1"])
        to_packages_dict(reqs)
        ```
        results in:
        ```
        {
            "pack_x": [<Requirement("pack_x[xtra]")>, <Requirement("pack_x!=2.1")>],
            "pack_y": [<Requirement("pack_y>=1.0")>]
        }
        ```

    Args:
        requirements (Iterable[Requirement]):
            A collection of requirements.

    Returns:
        Dict[str, List[Requirement]]:
            A dict that maps packge names to a list of specific requirements.
    """
    return {
        str(name): list(pkgs)
        for name, pkgs in itertools.groupby(requirements, lambda r: r.name)
    }


def from_packages_dict(
    packages: Mapping[Any, Iterable[Requirement]]
) -> Set[Requirement]:
    """Flattens any dict, where the values are collections of Requirements into a
    set of thoes requirements.

    Args:
        packages (Mapping[str, Iterable[Requirement]]):
            Any dict where the values are collections of requirements.
            For example the results from `to_package_dict`.

    Returns:
        Set[Requirement]:
            The set of unique requirments.
            Note that if the dictionary contains multiple versions
            (i.e different version constraints and/or extras) of the same requirement,
            the resulting set will contain them all too, as they are thechically not
            identical.
    """
    return set([req for _, req_list in packages.items() for req in req_list])


def resolve_requirements(
    requirements: Iterable[Union[str, Requirement]],
    enforce_redgrease: PackageOption = False,
) -> Set[Requirement]:
    """Attempts to resolve the given requirements, and that the desired version of
    redgrease is, or is NOT, included.

    Args:
        requirements (Iterable[Union[str, Requirement]]):
            Any iterable of requirements, encoded as either strings ('numpy',
            'passlib[bcrypt]==1.7.4') or a packaging.requirements.Requiremnt object.

        enforce_redgrease (PackageOption, optional):
            Indicates if, and potentially which version of, redgrease should be
            enforced as a requirement.

            It can take several optional types::

                - None : enforces that 'redgrease' is NOT in thequirements,
                    any redgrease requirements will be removed.

                - bool :
                    True, enforces latest 'redgrease' package on PyPi,
                    False, no enforcing. requirements are passed as-is,
                        with or without 'redgrease'

                - str :
                    a. Specific version. E.g. "1.2.3".
                    b. Version qualifier. E.g. ">=1.0.0"
                    c. Extras. E.g. "all" or "runtime".
                        Will enforce the latest version, with this/these extras
                    d. Custom source. E.g:
                        "redgrease[all]>=1.2.3"
                        "redgrease[runtime]@git+https://github.com/lyngon/redgrease.git@branch_xyz"

                - Version : behaves just as string version (a.)

                - Requirment : behaves just as string version (d.)

            Defaults to False.

    Returns:
        Set[Requirement]:
            Set of resulting Requirements.

    Raises:
        InvalidRequirement:
            If any of the requirements are not well-formed.
    """
    requirements_set = set(
        [packaging.requirements.Requirement(str(req)) for req in requirements]
        if requirements
        else []
    )

    packages = to_packages_dict(requirements_set)

    if enforce_redgrease is None:
        # enforce_redgrease is explicitly None, meaning we enforce that
        # redgrease should NOT be in ther requirements set
        # so we remove it
        if "redgrease" in packages:
            del packages["redgrease"]
        return from_packages_dict(packages)

    if isinstance(enforce_redgrease, bool):
        if enforce_redgrease and "redgrease" not in packages:
            # enforce_redgrease is True  (i.e: any version of redgrease)
            # and it is not currently in the requirement set
            # so we add the latest version from PyPi,
            packages["redgrease"] = [default_redgrease]
        return from_packages_dict(packages)

    if isinstance(enforce_redgrease, Requirement):
        # enforce_redgrease is a complete requirement
        # just enforce it and exit
        packages["redgrease"] = [enforce_redgrease]
        return from_packages_dict(packages)

    enforce_redgrease = str(enforce_redgrease)

    if enforce_redgrease[0] in ["=", ">", "<", "!"]:
        # enforce_redgrease is a qualifier / constraint string, e.g:
        # ">=1.0.0"
        redgrease_package = f"redgrease[runtime]{enforce_redgrease}"
    elif enforce_redgrease[0].isnumeric():
        # enforce_redgrease is a specific version string, e.g:
        # "1.2.3"
        redgrease_package = f"redgrease[runtime]=={enforce_redgrease}"
    elif "redgrease" not in enforce_redgrease:
        # enforce_redgrease does not contain 'redgrease'
        # so it is assumed to specify the 'extras'
        redgrease_package = "redgrease[{enforce_redgrease}]"
    else:
        # enforce_redgrease is some completely custom string, e.g:
        # "redgrease[runtime]@git+https://github.com/lyngon/redgrease.git@feature_xyz"
        redgrease_package = enforce_redgrease

    # Mostly to confirm that it is well formed
    redgrease_requirement = packaging.requirements.Requirement(redgrease_package)

    packages["redgrease"] = [redgrease_requirement]

    return from_packages_dict(packages)



# -*- coding: utf-8 -*-
"""
Basic syntactic sugar.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import logging
from typing import Any


class ReaderType:
    """Redis Gears Reader Types"""

    KeysReader = "KeysReader"
    """KeysReader"""

    KeysOnlyReader = "KeysOnlyReader"
    """KeysOnlyReader"""

    StreamReader = "StreamReader"
    """StreamReader"""

    PythonReader = "PythonReader"
    """PythonReader"""

    ShardsIDReader = "ShardsIDReader"
    """ShardsIDReader"""

    CommandReader = "CommandReader"
    """CommandReader"""


class TriggerMode:
    """Redis Geears Trigger modes for registered actions"""

    Async = "async"
    """Async"""

    AsyncLocal = "async_local"
    """AsyncLocal"""

    Sync = "sync"
    """Sync"""


class KeyType:
    """Redis Key Types"""

    String = "string"
    """String"""

    Hash = "hash"
    """Hash"""

    List = "list"
    """List"""

    Set = "set"
    """Set"""

    ZSet = "zset"
    """ZSet"""

    Stream = "stream"
    """Stream"""

    Module = "module"
    """Module"""

    _constructors = {
        "string": str,
        "hash": dict,
        "list": list,
        "set": set,
    }

    @staticmethod
    def of(python_type: Any) -> str:
        """Get the correspondig Redis key type string for a given Python type or value,
        if it exists.
        The only valid types are 'str', 'dict', 'list' and 'set'.

        Args:
            python_type (Any):
                A python type or any value value.

        Raises:
            ValueError:
                If there is no valid Redis key type for the Python type.

        Returns:
            str:
                The string representing the Redis key type.
        """
        if python_type is str or isinstance(python_type, str):
            return KeyType.String
        elif python_type is dict or isinstance(python_type, dict):
            return KeyType.Hash
        elif python_type is list or isinstance(python_type, list):
            return KeyType.List
        elif python_type is set or isinstance(python_type, set):
            return KeyType.Set
        else:
            raise ValueError(
                f"No corresponding Redis Key-Type for python type {python_type}"
            )

    @staticmethod
    def constructor(key_type: str):
        """Get the corresponting Python type / constructor, for a given Redis key type.

        Args:
            key_type (str):
                A Redis key type as str.

        Returns:
            Type:
                The Python constructor / type corresponding to the key type.
        """
        return KeyType._constructors[key_type]


class FailurePolicy:
    """Redis event handler failure policy"""

    Continue = "continue"
    """Continue"""

    Abort = "abort"
    """Abort"""

    Retry = "retry"
    """Retry"""


class LogLevel:
    """Redis Gears log levels"""

    Debug = "debug"
    """Debug"""

    Verbose = "verbose"
    """Verbose"""

    Notice = "notice"
    """Notice"""

    Warning = "warninig"
    """Warning"""

    @staticmethod
    def to_logging_level(rg_log_level):
        """Get the 'logging' log level corresponding to a Redis Gears log level.

        Args:
            rg_log_level ([type]):
                Reis Gears log level

        Returns:
            int:
                The corresponding default 'logging' log level.
        """
        if rg_log_level == LogLevel.Debug:
            return logging.DEBUG
        elif rg_log_level == LogLevel.Verbose:
            return logging.INFO
        elif rg_log_level == LogLevel.Notice:
            return logging.WARNING
        elif rg_log_level == LogLevel.Warning:
            return logging.ERROR  # Maybe a little unintuitive ;)
        else:
            return logging.INFO



# -*- coding: utf-8 -*-
"""
Gears function decorators
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from typing import Callable

import redgrease.runtime
import redgrease.sugar
from redgrease.gears import ClosedGearFunction
from redgrease.typing import Callback


def trigger(
    trigger: str,
    prefix: str = "*",
    convertToStr: bool = True,
    collect: bool = True,
    mode: str = redgrease.sugar.TriggerMode.Async,
    onRegistered: Callback = None,
    **kargs,
) -> Callable[[Callable], ClosedGearFunction]:
    """Decorator for creation of CommandReader + Tigger GearFunctions

    Args:
        trigger (str):
            The trigger string

        prefix (str, optional):
            Register prefix.
            Same as for the `register` operation.
            Defaults to "*".

        convertToStr (bool, optional):
            Convert the results to str.
            Same as for the `register` operation.
            Defaults to True.

        collect (bool, optional):
            Add a `collect' operation to the end of the function.
            Same as for the `register` operation.
            Defaults to True.

        mode (str, optional):
            The execution mode of the triggered function.
            Same as for the `register` operation.
            Defaults to redgrease.sugar.TriggerMode.Async.

        onRegistered (Callback, optional):
            A function callback thats called on each shard upon function registration.
            It is a good place to initialize non-serializable objects such as
            network connections.
            Same as for the `register` operation.
            Defaults to None.

    Returns:
        Callable[[Callable], ClosedGearFunction]:
            A ClosedGearFunction generator.
    """

    def gear(func):
        redgrease.runtime.GearsBuilder("CommandReader").map(
            lambda params: func(params[1:])
        ).register(
            prefix=prefix,
            convertToStr=convertToStr,
            collect=collect,
            mode=mode,
            onRegistered=onRegistered,
            trigger=trigger,
            **kargs,
        )

    return gear



# -*- coding: utf-8 -*-
"""
Utility and boilerplate functions, such as parsers, value transformers etc.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import ast
import functools
from enum import Enum
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from redgrease.typing import Constructor, Key, RedisType, T, Val


# Not a parser
class REnum(Enum):
    """Base Class for Redis-compatible enum values"""

    def __str__(self):
        return safe_str(self.value)

    def __bytes__(self):
        return to_redis_type(self.value)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    def __eq__(self, other):
        return str(self) == safe_str(other)

    def __hash__(self):
        return hash(str(self))


# Not a parser
def iteritems(d: Dict[Key, Val]) -> Iterable[Tuple[Key, Val]]:
    """Iterate items in a dict
    Just to le able to use CaseInsensitiveDict without modifications

    Args:
        d (Dict[Key, Val]):
            Dict to iterate

    Returns:
        Iterable[Tuple[Key, Val]]:
            Iterable of key-value tuples
    """
    return iter(d.items())


# Not a parser
class CaseInsensitiveDict(dict):
    """Case insensitive dict implementation.
    Assumes string keys only.
    Heavily influenced from redis.client.
    """

    def __init__(self, data):
        for k, v in iteritems(data):
            self[safe_str_upper(k)] = v

    def __contains__(self, k):
        return super(CaseInsensitiveDict, self).__contains__(safe_str_upper(k))

    def __delitem__(self, k):
        super(CaseInsensitiveDict, self).__delitem__(safe_str_upper(k))

    def __getitem__(self, k):
        return super(CaseInsensitiveDict, self).__getitem__(safe_str_upper(k))

    def get(self, k, default=None):
        return super(CaseInsensitiveDict, self).get(safe_str_upper(k), default)

    def __setitem__(self, k, v):
        super(CaseInsensitiveDict, self).__setitem__(safe_str_upper(k), v)

    def update(self, data):
        data = CaseInsensitiveDict(data)
        super(CaseInsensitiveDict, self).update(data)


def as_is(value: T) -> T:
    """Passthrough parser / identity function

    Args:
        value (T):
            Input value

    Returns:
        T:
            The value, unmodified.
    """
    return value


def str_if_bytes(value: T) -> Union[T, str]:
    """Parses byte values into a string, non-byte values passthrough unchanged
    Slightly modified from redis.utils, as it is not exported

    Args:
        value (T):
            Any serialized Redis value

    Returns:
        Union[T, str]:
            Either a string or the input unchanged
    """
    if isinstance(value, bytes):
        return value.decode("utf-8", errors="replace")
    return value


# Copied as is from redis.utils, as it is not exported
def safe_str(value: Any) -> str:
    """Parse anything to a string

    Args:
        value (Any):
            Input value

    Returns:
        str:
            String
    """
    if value is None:
        return ""
    return str(str_if_bytes(value))


def safe_str_upper(value: Any) -> str:
    """Parse anything to an uppercase string

    Args:
        value (Any):
            Input value

    Returns:
        str:
            Parsed uppercase string
    """
    return safe_str(value).upper()


def bool_ok(value: Any) -> bool:
    """Parse redis response as bool, such that:
    'Ok' => True
    Anything else => False
    Same name as in redis.client but slightly different implementation.
    Should be better for long non-Ok replies, e.g. images, erroneously passed to it

    Args:
        value (Any):
            Input value

    Returns:
        bool:
            Parsed boolean
    """
    if isinstance(value, (bytes, str)):
        return safe_str_upper(value[:2]) == "OK"
    else:
        return False


def safe_bool(input: Any) -> bool:
    """Parse a bool, slightly more accepting
    allowing for literal "True"/"False", integer 0/1,
    as well as "Ok" and "yes"/"no" values

    Args:
        input (Any):
            Input value

    Returns:
        bool:
            Parsed boolean
    """
    if isinstance(input, (bool, int)):
        return bool(input)
    if isinstance(input, (bytes, str)):
        return any(
            [
                safe_str_upper(input[:4]).startswith(true_val)
                for true_val in ["TRUE", "OK", "1", "Y", "YES"]
            ]
        )
    return False


# Not a parser
def to_int_if_bool(value: Any) -> Union[int, Any]:
    """Transforms any boolean into an integer
    As booleans are not natively supported as a separate datatype in Redis
    True => 1
    False => 0

    Args:
        value (Union[bool,Any]):
            A boolean value

    Returns:
        Union[int, Any]:
            Integer reprepresentataion of the bool
    """
    return int(value) if isinstance(value, bool) else value


# Not a parser
def to_redis_type(value: Any) -> RedisType:
    """Attempts to serialize a value to a Redis-native type.
    I.e. either: bytes, int or float
    It will serialze most primitive types (str, bool, int, float),
    as well as any complex type that implemens __bytes__ method

    Args:
        value (Any):
            Value to serialize for Redis

    Returns:
        RedisType:
            A serialized version
    """
    if value is None:
        return bytes()

    if isinstance(value, bool):
        return to_int_if_bool(value)

    if isinstance(value, str):
        return value.encode()

    if hasattr(value, "__bytes__"):
        value.__bytes__()

    if isinstance(value, (bytes, int, float)):
        return value

    return to_redis_type(str(value))


def to_bytes(value: Any) -> bytes:
    compat_val = to_redis_type(value)
    if isinstance(compat_val, bytes):
        return compat_val
    else:
        return str(compat_val).encode()


# Not a parser
def to_list(
    mapping: Optional[Dict[Key, Val]],
    key_transform: Callable[[Key], Any] = str_if_bytes,
    val_transform: Callable[[Val], Any] = to_redis_type,
) -> List:
    """Flattens a Dict to a list consisting of the of keys and values altertnating.
    This is useful generating the arguments for Redis commands.

    Args:
        mapping (Dict[Key, Val]):
            The dict to flatten.

        key_transform (Callable[[Key], Any], optional):
            Transformation function for the keys.
            Defaults to 'str_if_bytes'

        val_transform (Callable[[Val], Any], optional):
            Transformation function for the values.
            Defaults to 'to_redis_type'.

    Returns:
        List:
            Flattened list of the transformed keys and values
    """
    if mapping is None:
        return []
    return list(
        [
            item
            for key, value in mapping.items()
            for item in (key_transform(key), val_transform(value))
        ]
    )


def transform(
    value: Any,
    constuctor: Union[Constructor[T], Dict[Key, Constructor[T]]],
    key: Key = None,
) -> T:
    """Applies a transformation to a value.
    The tranformation fuction could either be passed directly or in a dictionary along
    with a key to look it up.
    This is mostly only useful as a helper function for constructors of composite types,
    where the value may need to be transformed differently depending on the field/key.

    Args:
        value (Any):
            Value to transform

        constuctor (Union[Constructor[T], Dict[Any, Constructor[T]]]):
            Transformation function or a dict of transformation functions.
            If a dict is used, the key argument is used to look up the transformation.
            If the key is not present in the dict, but there is a 'None'-key mapping
            to a functsion, this function will be used as a default.
            Otherwise, the value will be returned untransformed.

        key (Key):
            key to use to look up the appropriate transforpation

    Returns:
        T: Transformed value
    """
    if constuctor is None:
        return value  # type: ignore

    try:
        if isinstance(constuctor, dict):
            if key in constuctor:
                return constuctor[key](value)  # type: ignore
            elif None in constuctor:
                return constuctor[None](value)  # type: ignore
            else:
                return value
        else:
            return constuctor(value)  # type: ignore
    except (TypeError, ValueError):
        # if for some reason the value can't be coerced, just use
        # the raw value
        return value


def to_dict(
    items: Iterable,
    keys: Iterable = None,
    key_transform: Union[Constructor[Key], Dict[Any, Constructor[Key]]] = None,
    val_transform: Union[Constructor[Val], Dict[Key, Constructor[Val]]] = None,
) -> Dict[Key, Val]:
    """Folds an iterable of values into a dict.
    This is useful for parsing Redis' list responseses into a more manageable structure.
    It can be used on lists of following different structures:

    - Alternating unnamed Key and values, i.e:
    [key_1, value_1, key_2, value_2, ... ]
        eg:
        - to_dict(["foo", 42, 13, 37]) == {"foo": 42, 13: 37}
        - to_dict(["foo", 42, 13, 37], key_transform=str) == {"foo": 42, "13": 37}
        - to_dict(["foo", 42, 13, 37], val_transform=str) == {"foo": "42", 13: "37"}
        - to_dict(["foo", 42, 13, 37], val_transform={"foo":int, 13:float})
            == {"foo": 42, 13: 37.0}
        - to_dict(
            ["foo", 42, 13, 37],
            key_transform=str,
            val_transform={"foo":int, "13":float}
        ) == {"foo": 42, "13": 37.0}


    Args:
        items (Iterable):
            Iterable to "fold" into a dict

        key_transform (Union[Constructor[Key], Dict[Any, Constructor[Key]]], optional):
            Transformation function / type / constructor to apply to keys.
            It can either be a callable, which is then applied to all keys.
            Altertnatively, it can be a mapping from 'raw' key to a specific transform
            for that key
            Defaults to None (No key transformation).

        val_transform (Union[Constructor[Val], Dict[Key, Constructor[Val]]], optional):
            Transformation function / type / constructor to apply to values.
            It can either be a callable, which is then applied to all values.
            Altertnatively, it can be a mapping from (transformed) key to a specific
            transform for the value of that key
            Defaults to None (No value transformation).

    Returns:
        Dict[Key, Val]:
            Folded dictionary
    """

    if items is None:
        return {}  # type: ignore

    if key_transform is None:
        key_transform = as_is

    if val_transform is None:
        val_transform = as_is

    if keys:
        kv_pairs = zip(keys, items)
    else:
        it = iter(items)
        kv_pairs = zip(it, it)

    result = {}
    for key, value in kv_pairs:
        key = transform(key, key_transform, key)
        value = transform(value, val_transform, key)  # type: ignore
        result[key] = value

    return result


def to_kwargs(items: Iterable) -> Dict[str, Any]:
    """Folds an iterable of values into a 'kwargs-compatible' dict.
    This is useful for cunstucting objects from  Redis' list responseses, by means
    of an intermediate kwargs dict that can be passed to for example a constructor.
    It behaves exactly as 'to_dict' but enforces keys to be parsed to strings.

    - Alternating unnamed Key and values, i.e:
    [key_1, value_1, key_2, value_2, ... ]
        eg:
            input:  ["foo", 42, 13, 37]
            output: {"foo": 42, "13": 37}

    Args:
        items (Iterable):
            Iterable to "fold" into a dict

    Returns:
        Dict[str, Any]:
            Folded dictionary
    """
    return to_dict(items, key_transform=safe_str)


# TODO: Should maybe be renamed to list_of or parse_list or sometiong
def list_parser(item_parser: Constructor[T]) -> Callable[[Iterable], List[T]]:
    """Creates a list parser for lists of objects created with a given constructor.
    E.g:

    parser = list_parser(bool)
    parser(['', 1, None])

    => [False, True, False]

    Args:
        item_parser (Constructor[T]):
            The constructor to apply to each element.

    Returns:
        Callable[[Iterable[Any]], List[T]]:
            Function that takes maps the constructor on to the iterable and returns
            the result as a list.
    """

    def parser(input_list):
        if isinstance(input_list, bytes):
            input_list = safe_str(input_list)
        if isinstance(input_list, str):
            input_list = list(ast.literal_eval(input_list))

        return list(map(item_parser, input_list))

    return parser


# is this used anywhere?
list_of_str = list_parser(safe_str)


def dict_of(
    constructors: Dict[Key, Constructor[Any]]
) -> Callable[[Iterable, Iterable[Key]], Dict[Key, Any]]:
    """Creates a parser that parses a list of values to a dict,
    according to a dict of named constructors.

    The generated parser takes both the iterable of values to parse,
    as well as, an equally long, iterable of names/keys to to use to lookup the
    corresponding parser/constructor in the constructor lookup dict.

    The parser for the Nth value is using the parser found by looking up the Nth
    name/key in the key list in the lookup dict.
    This key is also used as the key in the resulting dict.

    E.g:
    parser = dict_of({"b":bool, "i":int, "s":str, "f":float})
    parser([0,1,0,1], ["b","f","s","i"])

    => {"b":False, "f":1.0, "i":1, "s":"0"}

    Args:
        constructors (Dict[str, Constructor[Any]]):
            Dict of named constructors

    Returns:
        Callable[..., Dict[str, Any]]:
            Dict parser
    """

    def parser(values: Iterable, keys: Iterable[Key]):
        return {
            key: transform(value, constructors, key)
            for (key, value) in zip(keys, values)
        }

    return parser


class Record:
    """Class repredenting a Redis Record, as generated by KeysReader.

    Attributes:
        key (str):
            The name of the Redis key.

        value (Any):
            The value corresponting to the key. `None` if deleted.

        type (str):
            The core Redis type. Either 'string', 'hash', 'list', 'set', 'zset' or
            'stream'.

        event (str):
            The event that triggered the execution.
            (`None` if the execution was created via the run function.)
    """

    def __init__(
        self,
        key: str,
        value: Any = None,
        type: str = None,
        event: str = None,
    ):

        self.key = key
        self.value = value
        self.type = type
        self.event = event


def record(rec: Union[str, dict]) -> Record:
    """Create a Record

    Args:
        rec (Union[str, dict]):
            the value to parse. Either a string (key only) or a dict with at minimum
            the key `key` present and optionally any of the keys `value`, `type`
            and/or `event`.

    Returns:
        Record:
            Parsed Record object.
    """
    if isinstance(rec, Record):
        return rec

    if isinstance(rec, dict):
        return Record(**rec)

    if isinstance(rec, str):
        return Record(rec)


def compose(*function: Callable) -> Callable:
    """Compose functions.
    I.e::

        ``lambda x: f(g(x))``
        can be written:
        ``compose(f, g)``

    Args:
        *function (Callable):
            Any number of functions to compose together.
            Output type of function N must be the input type of function N+1.

    Returns:
        Callable:
            A composed function with input type same as the firs function, and output
            type same as the last function.
    """
    return functools.reduce(lambda f, g: lambda x: f(g(x)), function, lambda x: x)



# -*- coding: utf-8 -*-
"""
Redgrease's (overloaded) variants of the symbols loaded per default into the top level
namespace of Gear functions in the Redis server Python runtime.
"""
__author__ = "Anders Åström"
__contact__ = "anders@lyngon.com"
__copyright__ = "2021, Lyngon Pte. Ltd."
__licence__ = """The MIT License
Copyright © 2021 Lyngon Pte. Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of this
 software and associated documentation files (the “Software”), to deal in the Software
 without restriction, including without limitation the rights to use, copy, modify,
 merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
 permit persons to whom the Software is furnished to do so, subject to the following
 conditions:

The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import redgrease.gears
import redgrease.sugar as sugar


class GearsBuilder(redgrease.gears.PartialGearFunction):
    """The GearsBuilder class is imported to the runtime's environment by default.

    It exposes the functionality of the function's context builder.
    """

    def __init__(
        self,
        reader: str = sugar.ReaderType.KeysReader,
        defaultArg: str = "*",
        desc: str = None,
        *args,
        **kwargs,
    ):
        """Gear function / process factory

        Args:
            reader (str, optional):
                Input records reader
                Defining where the input to the gear will come from.
                One of::
                    - 'KeysReader'
                    - 'KeysOnlyReader'
                    - 'StreamReader'
                    - 'PythonReader'
                    - 'ShardsReader'
                    - 'CommandReader'
                Defaults to 'KeysReader'.

            defaultArg (str, optional):
                Additional arguments to the reader.
                These are usually a key's name, prefix, glob-like or a regular
                expression.
                Its use depends on the function's reader type and action.
                Defaults to '*'.

            desc (str, optional):
                An optional description.
                Defaults to None.
        """
        reader_op = redgrease.gears.Reader(reader, defaultArg, desc, *args, **kwargs)
        super().__init__(
            operation=reader_op,
            input_function=None,
        )


GB = GearsBuilder
"""Convnenience shorthand for GearsBuilder."""


# # Suppress warnings for missing redisgears packagae
# # As this package only lives on the Redis Gears server
# pyright: reportMissingImports=false
class atomic:
    """The atomic() Python context is imported to the runtime's environment by default.

    The context ensures that all operations in it are executed atomically by blocking
    he main Redis process.
    """

    def __init__(self):
        from redisgears import atomicCtx as redisAtomic

        self.atomic = redisAtomic()

    def __enter__(self):
        self.atomic.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self.atomic.__exit__(*args, **kwargs)


def execute(command: str, *args) -> bytes:
    """Execute an arbitrary Redis command.

    Args:
        command (str):
            The commant to execute

    Returns:
        bytes:
            Raw command response
    """
    from redisgears import executeCommand as redisExecute

    return redisExecute(command, *args)


def hashtag() -> bytes:
    """Returns a hashtag that maps to the lowest hash slot served by the local
    engine's shard. Put differently, it is useful as a hashtag for partitioning
    in a cluster.

    Returns:
        str:
            A hastag that maps to the lowest hash slot served by the local engine.
    """
    from redisgears import getMyHashTag as redisHashtag

    return redisHashtag()


def log(message: str, level: str = sugar.LogLevel.Notice):
    """Print a message to Redis' log.

    Args:
        message (str):
            The message to output

        level (str, optional):
            Message loglevel. Either::

                - 'debug'
                - 'verbose'
                - 'notice'
                - 'wartning'

            Defaults to 'notice'.
    """
    from redisgears import log as redisLog

    return redisLog(str(message), level=level)


def configGet(key: str) -> bytes:
    """Fetches the current value of a RedisGears configuration option.

    Args:
        key (str):
            The configuration option key
    """
    from __main__ import configGet as redisConfigGet

    return redisConfigGet(key)


def gearsConfigGet(key: str, default=None) -> bytes:
    """Fetches the current value of a RedisGears configuration option and returns a
    default value if that key does not exist.

    Args:
        key (str):
            The configuration option key.

        default ([type], optional):
            A default value.
            Defaults to None.
    """
    from __main__ import gearsConfigGet as redisGearsConfigGet

    return redisGearsConfigGet(key)


def run(
    function: redgrease.gears.GearFunction,
    builder: GearsBuilder,
):
    """Transforms a RedGrease GearFunction into a native Gears function.

    Note: This function is specific to RedGrease and is NOT a standard Redis Gears
    runtime function.

    Args:
        function (redgrease.gears.GearFunction):
            A GearsFunction, as created with RedGrease constructs.

        builder (GearsBuilder):
            The Redis Gears native GearsBuilder class.

    Returns:
        GearsBuilder:
            Redis Gears native GearsBuilder object, reconstructed from the input.
    """
    if function.input_function:
        input_builder = run(function.input_function, builder)
    else:
        input_builder = builder

    return function.operation.add_to(input_builder)
